@@ -1,3 +1,5 @@
+"""S3 bucket helper utilities."""
+
 import json
 import logging
 import mimetypes
@@ -16,10 +18,15 @@ from envidat.s3 import exceptions
 from envidat.utils import get_url
 
 log = logging.getLogger(__name__)
+if "DEBUG_BOTO" not in os.environ:
+    logging.getLogger("boto3").setLevel(logging.CRITICAL)
+    logging.getLogger("botocore").setLevel(logging.CRITICAL)
+    logging.getLogger("s3transfer").setLevel(logging.CRITICAL)
 
 
 class Bucket:
     """Class to handle S3 bucket transactions.
+
     Handles boto3 exceptions with custom exception classes.
     """
 
@@ -42,7 +49,6 @@ class Bucket:
             is_new (bool): If true, creates a new bucket.
             is_public (bool): If true, makes the bucket public on creation.
         """
-
         if None in [Bucket._AWS_ACCESS_KEY_ID, Bucket._AWS_SECRET_ACCESS_KEY]:
             missing_vars = {"AWS_ACCESS_KEY", "AWS_SECRET_KEY"} - set(os.environ)
             if not missing_vars:
@@ -107,7 +113,6 @@ class Bucket:
     @staticmethod
     def get_boto3_resource() -> NoReturn:
         """Configure boto3 resource object."""
-
         log.debug("Accessing boto3 resource.")
         return boto3.resource(
             "s3",
@@ -121,7 +126,6 @@ class Bucket:
     @staticmethod
     def get_boto3_client() -> NoReturn:
         """Cofigure boto3 client object."""
-
         log.debug("Accessing boto3 client.")
         return boto3.client(
             "s3",
@@ -134,6 +138,7 @@ class Bucket:
 
     def _handle_boto3_client_error(self, e: ClientError, key: str = None) -> NoReturn:
         """Handle boto3 ClientError.
+
         The exception type returned from the server is nested here.
         Refer to exceptions.py
 
@@ -166,7 +171,6 @@ class Bucket:
             is_dir (bool): True if path is a directory.
                 Defaults to False.
         """
-
         msg = (
             f"Referenced {'directory' if is_dir else 'file'} "
             f"not found on disk: {file_path}"
@@ -181,7 +185,6 @@ class Bucket:
             param_name (str): The parameter name.
             value (str): The parameter value.
         """
-
         if value is None:
             msg = f"A value must be set for parameter {param_name}"
         else:
@@ -191,12 +194,12 @@ class Bucket:
 
     def create(self) -> "boto3.resource.Bucket":
         """Create the S3 bucket on the endpoint.
+
         Method may be called directly to manipulate the boto3 Bucket object.
 
         Returns:
             "boto3.resource.Bucket": A boto3 S3 Bucket object.
         """
-
         resource = Bucket.get_boto3_resource()
 
         try:
@@ -226,6 +229,7 @@ class Bucket:
         decode: bool = False,
     ) -> (Any, dict):
         """Get an object from the bucket into a memory object.
+
         Defaults to utf-8 decode, unless specified.
 
         Args:
@@ -238,9 +242,8 @@ class Bucket:
         Returns:
             tuple: (data, S3 Metadata dict).
         """
-
         resource = Bucket.get_boto3_resource()
-        s3_object = resource.Object(self.bucket_name, key)
+        s3_object = resource.Object(self.bucket_name, key.lstrip("/"))
 
         try:
             log.info(f"Getting S3 object with key {key}")
@@ -284,9 +287,8 @@ class Bucket:
         Returns:
             dict: Response dictionary from S3.
         """
-
         resource = Bucket.get_boto3_resource()
-        s3_object = resource.Object(self.bucket_name, key)
+        s3_object = resource.Object(self.bucket_name, key.lstrip("/"))
 
         try:
             log.info(
@@ -306,7 +308,7 @@ class Bucket:
         except ClientError as e:
             self._handle_boto3_client_error(e, key=key)
 
-    def delete(self, key: str) -> dict:
+    def delete_file(self, key: str) -> dict:
         """Delete specified object of a given key.
 
         Args:
@@ -315,7 +317,6 @@ class Bucket:
         Returns:
             dict: Response dictionary from S3.
         """
-
         client = Bucket.get_boto3_client()
 
         try:
@@ -328,6 +329,7 @@ class Bucket:
 
     def upload_file(self, key: str, local_filepath: Union[str, Path]) -> bool:
         """Upload a local file to the bucket.
+
         Transparently manages multipart uploads.
 
         Args:
@@ -337,9 +339,8 @@ class Bucket:
         Returns:
             bool: True if success, False is failure.
         """
-
         resource = Bucket.get_boto3_resource()
-        s3_object = resource.Object(self.bucket_name, key)
+        s3_object = resource.Object(self.bucket_name, key.lstrip("/"))
 
         file_path = Path(local_filepath).resolve()
         if not file_path.is_file():
@@ -366,6 +367,7 @@ class Bucket:
 
     def download_file(self, key: str, local_filepath: Union[str, Path]) -> bool:
         """Download S3 object to a local file.
+
         Transparently manages multipart downloads.
 
         Args:
@@ -376,9 +378,8 @@ class Bucket:
         Returns:
             bool: True if success, False is failure.
         """
-
         resource = Bucket.get_boto3_resource()
-        s3_object = resource.Object(self.bucket_name, key)
+        s3_object = resource.Object(self.bucket_name, key.lstrip("/"))
 
         file_path = Path(local_filepath).resolve()
         if not file_path.parent.is_dir():
@@ -400,6 +401,7 @@ class Bucket:
 
     def transfer(self, source_key: str, dest_bucket: str, dest_key: str = None) -> bool:
         """Fast efficient transfer bucket --> bucket using TransferManager.
+
         This function avoids downloading to memory and uses the underlying
         operations that aws-cli uses to transfer.
 
@@ -412,7 +414,6 @@ class Bucket:
         Returns:
             bool: True if success, False is failure.
         """
-
         client = Bucket.get_boto3_client()
 
         if dest_key is None:
@@ -452,9 +453,8 @@ class Bucket:
         """Get a list of all objects in the bucket.
 
         Returns:
-            list: List of s3.ObjectSummary dicts, containing object metadata.
+            list: All keys in the bucket.
         """
-
         resource = Bucket.get_boto3_resource()
 
         try:
@@ -464,7 +464,7 @@ class Bucket:
             log.debug("Listing all objects in bucket")
             objects = bucket.objects.all()
 
-            file_names = [os.path.splitext(file.key)[0] for file in objects]
+            file_names = [file.key for file in objects]
             log.info(
                 f"Returned {len(file_names)} objects from "
                 f"bucket named {self.bucket_name}"
@@ -483,6 +483,7 @@ class Bucket:
         names_only: bool = False,
     ) -> list:
         """Get a list of all objects in a specific directory (s3 path).
+
         Returns up to a max of 1000 values.
 
         Args:
@@ -499,7 +500,6 @@ class Bucket:
         Returns:
             list: List of s3.ObjectSummary dicts, containing object metadata.
         """
-
         resource = Bucket.get_boto3_resource()
 
         if path:
@@ -570,11 +570,10 @@ class Bucket:
             dict: key:value pair of s3_key:download_status.
                 download_status True if downloaded, False if failed.
         """
-
         status_dict = {}
 
         local_dir_path = Path(local_dir)
-        log.debug(f"Top level directory to download to: {local_dir_path}")
+        log.debug(f"Downloading S3 directory to: {str(local_dir)}")
 
         s3_keys = self.list_dir(path=s3_path, recursive=True, file_type=file_type)
 
@@ -589,52 +588,111 @@ class Bucket:
 
         return status_dict
 
-    def upload_dir(
+    def download_all(
         self,
-        s3_path: str,
         local_dir: Union[str, Path],
         file_type: str = "",
     ) -> bool:
         """
-        Upload an entire local directory to a bucket path.
+        Download an entire S3 bucket, including subpaths, to a local directory.
+
+        Args:
+            local_dir (Union[str, Path]): Directory to download files into.
+            file_type (str): Download files with extension only, e.g. txt.
+
+        Returns:
+            dict: key:value pair of s3_key:download_status.
+                download_status True if downloaded, False if failed.
+        """
+        status_dict = self.download_dir("", local_dir, file_type)
+
+        return status_dict
+
+    def upload_dir(
+        self,
+        local_dir: Union[str, Path],
+        s3_path: str = "/",
+        file_type: str = "",
+        contents_only: bool = False,
+    ) -> bool:
+        """
+        Upload the content of a local directory to a bucket path.
 
         Args:
-            s3_path (str): The path within the bucket to upload to.
             local_dir (Union[str, Path]): Directory to upload files from.
-            file_type (str): Upload files with extension only, e.g. txt.
+            s3_path (str, optional): The path within the bucket to upload to.
+                If omitted, the bucket root is used.
+            file_type (str, optional): Upload files with extension only, e.g. txt.
+            contents_only (bool): Used to copy only the directory contents to the
+                specified path, not the directory itself.
 
         Returns:
             dict: key:value pair of file_name:upload_status.
                 upload_status True if uploaded, False if failed.
         """
-
         status_dict = {}
 
         local_dir_path = Path(local_dir).resolve()
-        log.debug(f"Full directory path to upload: {local_dir_path}")
+        log.debug(f"Directory to upload: {local_dir_path}")
 
-        parent_parts = local_dir_path.parent.parts
-        log.debug(f"Part directory parts: {parent_parts}")
-        num_parents = len(parent_parts)
-
-        all_subdirs = list(local_dir_path.glob("**"))
+        all_subdirs = local_dir_path.glob("**")
 
         for dir_path in all_subdirs:
-            log.debug(f"Searching for files in directory: {dir_path}")
 
+            log.debug(f"Searching for files in directory: {dir_path}")
             file_names = dir_path.glob(f"*{('.' + file_type) if file_type else ''}")
 
             # Only return valid files
             file_names = [f for f in file_names if f.is_file()]
             log.debug(f"Files found: {list(file_names)}")
 
-            for i, file_name in enumerate(file_names):
-                s3_key = str(Path(s3_path) / Path(*file_name.parts[num_parents:]))
+            for _, file_name in enumerate(file_names):
+                s3_key = str(
+                    Path(s3_path)
+                    / file_name.relative_to(
+                        local_dir_path if contents_only else local_dir_path.parent
+                    )
+                )
                 log.debug(f"S3 key to upload: {s3_key}")
                 status_dict[str(file_name)] = self.upload_file(s3_key, file_name)
 
         return status_dict
 
+    def delete_dir(
+        self,
+        s3_path: str,
+        file_type: str = "",
+    ) -> bool:
+        """
+        Delete an entire S3 path, including subpaths.
+
+        USE WITH CAUTION!
+
+        Args:
+            s3_path (str): The path within the bucket to delete.
+            file_type (str): Delete files with extension only, e.g. txt.
+
+        Returns:
+            dict: key:value pair of s3_key:deletion_status.
+                deletion_status True if deleted, False if failed.
+        """
+        status_dict = {}
+
+        s3_keys = self.list_dir(path=s3_path, recursive=True, file_type=file_type)
+
+        for key in s3_keys:
+            log.info(f"Deleting key: {key}")
+            response = self.delete_file(key)
+
+            if response["ResponseMetadata"]["HTTPStatusCode"] == 204:
+                log.debug("Key successfully deleted.")
+                status_dict[key] = True
+            else:
+                log.debug("Key deletion failed.")
+                status_dict[key] = False
+
+        return status_dict
+
     def check_file_exists(self, key: str) -> bool:
         """Check an object exists in the bucket.
 
@@ -644,12 +702,11 @@ class Bucket:
         Returns:
             bool: True if exists, False if not.
         """
-
         client = Bucket.get_boto3_client()
 
         try:
             log.info(f"Retrieving S3 object metadata with key: {key}")
-            response = client.head_object(Bucket=self.bucket_name, Key=key)
+            response = client.head_object(Bucket=self.bucket_name, Key=key.lstrip("/"))
 
             if response["ResponseMetadata"]["HTTPStatusCode"] == 200:
                 return True
@@ -666,6 +723,41 @@ class Bucket:
 
             self._handle_boto3_client_error(e, key=key)
 
+    def rename_file(self, key: str, dest_key: str) -> bool:
+        """Rename a file in a bucket, i.e. move then delete source.
+
+        Args:
+            key (str): The key, i.e. path within the bucket.
+            dest_key (str): The key destination to move to.
+
+        Returns:
+            bool: True if success, False if skipped or failure.
+        """
+        resource = Bucket.get_boto3_resource()
+
+        try:
+            if not self.check_file_exists(key):
+                log.info("File does not exist, cannot rename.")
+                return False
+
+            log.info(f"Copying file: {key} to destination: {dest_key}")
+            response = resource.Object(self.bucket_name, dest_key).copy_from(
+                CopySource={"Bucket": self.bucket_name, "Key": key}
+            )
+            log.debug(f"Copy file response: {response}")
+
+            if response["ResponseMetadata"]["HTTPStatusCode"] == 200:
+                log.info(f"Deleting original file: {key}")
+                resource.Object(self.bucket_name, key).delete()
+                return True
+
+            else:
+                log.error(f"Copying file {key} failed. Aborting deletion")
+                return False
+
+        except ClientError as e:
+            self._handle_boto3_client_error(e, key=key)
+
     def clean_multiparts(self) -> bool:
         """
         Clean up failed multipart uploads in a bucket.
@@ -674,7 +766,6 @@ class Bucket:
             dict: key:value pair of s3_multipart_key:clean_status.
                 clean_status True if removed, False if failed.
         """
-
         status_dict = {}
         success_counter = 0
         failure_counter = 0
@@ -737,7 +828,6 @@ class Bucket:
         Returns:
             bool: True if success, False is failure.
         """
-
         client = Bucket.get_boto3_client()
 
         try:
@@ -799,7 +889,6 @@ class Bucket:
         Returns:
             dict: Response dictionary from index file upload.
         """
-
         if isinstance(file_list, str):
             log.debug(f"Converting string file_list into list: {file_list}")
             file_list = [file_list]
@@ -856,7 +945,6 @@ class Bucket:
         Returns:
             dict: Response dictionary containing CORS config.
         """
-
         client = Bucket.get_boto3_client()
 
         try:
@@ -885,7 +973,6 @@ class Bucket:
         Returns:
             bool: True if success, False is failure.
         """
-
         if allow_all is False and origins is None:
             log.debug("No origins provided and allow_all not set. Skipping")
             self._raise_parameter_error("origins", origins)
